<?php

namespace Drupal\media_dupe_detect\Form;

use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\InvokeCommand;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\PageCache\ResponsePolicy\KillSwitch;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Url;
use Drupal\media\MediaInterface;
use Drupal\media_library\Form\FileUploadForm;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form for uploading media items with duplicate detection.
 *
 * @phpstan-ignore-next-line
 */
class DupeDetectImageUploadForm extends FileUploadForm {

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The page cache kill switch.
   *
   * @var \Drupal\Core\PageCache\ResponsePolicy\KillSwitch
   */
  protected $killSwitch;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The file URL generator.
   *
   * @var \Drupal\Core\File\FileUrlGeneratorInterface
   */
  protected $fileUrlGenerator;

  /**
   * Adds additional required services.
   */
  public function addServices(
    MessengerInterface $messenger,
    ConfigFactoryInterface $config_factory,
    FileUrlGeneratorInterface $file_url_generator,
    Connection $database,
    AccountProxyInterface $current_user,
    KillSwitch $kill_switch,
  ) {
    $this->messenger = $messenger;
    $this->configFactory = $config_factory;
    $this->fileUrlGenerator = $file_url_generator;
    $this->database = $database;
    $this->currentUser = $current_user;
    $this->killSwitch = $kill_switch;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = new self(
      $container->get('entity_type.manager'),
      $container->get('media_library.ui_builder'),
      $container->get('element_info'),
      $container->get('renderer'),
      $container->get('file_system'),
      $container->get('media_library.opener_resolver'),
      $container->get('file.usage'),
      $container->get('file.repository')
    );

    $instance->addServices(
      $container->get('messenger'),
      $container->get('config.factory'),
      $container->get('file_url_generator'),
      $container->get('database'),
      $container->get('current_user'),
      $container->get('page_cache_kill_switch')
    );

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    if ($form_state->get('redirect_to_duplicate')) {
      $duplicate_id = $form_state->get('duplicate_media_id');

      // Get current request parameters.
      $current_request = $this->getRequest();

      // Build media library URL with necessary parameters.
      $params = [
        'media_library_opener_id' => $current_request->query->get('media_library_opener_id'),
        'media_library_allowed_types' => $current_request->query->get('media_library_allowed_types'),
        'media_library_selected_type' => $current_request->query->get('media_library_selected_type'),
        'media_library_remaining' => $current_request->query->get('media_library_remaining', -1),
        'media_library_opener_parameters' => $current_request->query->get('media_library_opener_parameters'),

        // Pre-select our existing media item.
        // @todo this is not working as expected.
        'media_library_selected_ids' => [$duplicate_id],
      ];

      if ($hash = $current_request->query->get('hash')) {
        $params['hash'] = $hash;
      }

      // Create URL for the media library grid view.
      $url = Url::fromRoute('media_library.ui')
        ->setOption('query', $params);

      $form_state->setRedirectUrl($url);
      return;
    }

    parent::submitForm($form, $form_state);
  }

  /**
   * Check if a file with the same MD5 hash exists.
   */
  protected function findDuplicatesByHash($filepath) {
    $md5 = md5_file($filepath);
    if (!$md5) {
      return [];
    }

    $query = $this->database->select('file_managed', 'fm');
    $query->join('media__field_media_image', 'mfi', 'mfi.field_media_image_target_id = fm.fid');
    $query->join('media_field_data', 'mfd', 'mfd.mid = mfi.entity_id');

    $query->fields('mfd', ['mid', 'name'])
      ->fields('fm', ['fid', 'filename', 'uri'])
      ->condition('fm.md5', $md5)
      ->condition('fm.status', 1);

    return $query->execute()->fetchAll();
  }

  /**
   * {@inheritdoc}
   */
  protected function buildEntityFormElement(MediaInterface $media, array $form, FormStateInterface $form_state, $delta) {
    $element = parent::buildEntityFormElement($media, $form, $form_state, $delta);

    // Check if duplicate detection is enabled for this field.
    if (!$this->isDuplicateDetectionEnabled($form_state)) {
      return $element;
    }

    $source = $media->getSource();
    $source_field = $source->getSourceFieldDefinition($media->bundle->entity);
    $field_name = $source_field->getName();

    if ($media->hasField($field_name) && !$media->get($field_name)->isEmpty()) {
      $file = $media->get($field_name)->entity;
      if ($file) {
        $duplicates = $this->findDuplicatesByHash($file->getFileUri());

        if (!empty($duplicates)) {
          // Add our CSS and JS.
          $element['#attached']['library'][] = 'media_dupe_detect/duplicate_detection';

          // Get the first duplicate media entity to display.
          $first_duplicate = reset($duplicates);
          if ($existing_media = $this->entityTypeManager->getStorage('media')->load($first_duplicate->mid)) {
            if ($existing_media->access('view')) {
              // Add the duplicate detection elements inside the fields wrapper.
              $element['fields'] = [
                'duplicate_notice' => [
                  '#type' => 'container',
                  '#attributes' => ['class' => ['media-library-duplicate-check']],
                  '#weight' => -101,
                  'message' => [
                    '#markup' => $this->t('A similar image already exists in the media library:'),
                    '#prefix' => '<div class="duplicate-check-message">',
                    '#suffix' => '</div>',
                  ],
                  'item' => [
                    '#type' => 'container',
                    '#attributes' => ['class' => ['duplicate-item-wrapper']],
                    'thumbnail' => [
                      '#theme' => 'image_style',
                      '#style_name' => 'thumbnail',
                      '#uri' => $existing_media->get($field_name)->entity->getFileUri(),
                    ],
                    'info' => [
                      '#type' => 'container',
                      '#attributes' => ['class' => ['duplicate-info']],
                      'name' => [
                        '#markup' => $existing_media->label(),
                        '#prefix' => '<div class="name">',
                        '#suffix' => '</div>',
                      ],
                      'filename' => [
                        '#markup' => $first_duplicate->filename,
                        '#prefix' => '<div class="filename">',
                        '#suffix' => '</div>',
                      ],
                    ],
                  ],
                  'use_existing' => [
                    '#type' => 'submit',
                    '#value' => $this->t('Use Existing'),
                    '#ajax' => [
                      'callback' => '::useExistingCallback',
                      'wrapper' => 'media-library-add-form-wrapper',
                      'message' => $this->t(
                        'Returning to library to select @name.',
                        ['@name' => $first_duplicate->filename]
                      ),
                    ],
                    '#submit' => ['::useExistingSubmit'],
                    '#limit_validation_errors' => [],
                    '#attributes' => [
                      'class' => ['use-existing-button'],
                      'data-filename' => $first_duplicate->filename,
                      'data-mid' => $first_duplicate->mid,
                    ],
                  ],
                ],
              ];

              // Display acknowledgement checkbox, if user has permission.
              $element['fields']['duplicate_acknowledge'] = [
                '#type' => 'checkbox',
                '#title' => $this->t('I want to create a new media item instead of using the existing one'),
                '#weight' => -100,
                '#attributes' => ['class' => ['duplicate-acknowledge']],
                '#disabled' => !$this->currentUser->hasPermission('bypass media duplicate upload prevention'),
              ];

              if ($this->currentUser->hasPermission('bypass media duplicate upload prevention')) {
                // Add the original fields back to the form.
                $element['fields'] = $element['fields'] + parent::buildEntityFormElement($media, $form, $form_state, $delta)['fields'];
              }
            }
          }
        }
      }
    }

    return $element;
  }

  /**
   * Submit handler for the Use Existing button.
   */
  public function useExistingSubmit(array $form, FormStateInterface $form_state) {
    // Get the filename and mid from the button.
    $triggering_element = $form_state->getTriggeringElement();
    $filename = $triggering_element['#attributes']['data-filename'];
    $mid = $triggering_element['#attributes']['data-mid'];

    // Store the filename and mid in form state for the callback.
    $form_state->set('filter_filename', $filename);
    $form_state->set('filter_mid', $mid);

    // Clean up the temporary file.
    if ($temp_file = $form_state->get('temporary_file')) {
      if ($file = $this->entityTypeManager->getStorage('file')->load($temp_file['fid'])) {
        $file->delete();
      }
      if (file_exists($temp_file['uri'])) {
        $this->fileSystem->delete($temp_file['uri']);
      }
    }

    $form_state->setRebuild();
  }

  /**
   * AJAX callback for the Use Existing button.
   */
  public function useExistingCallback(array &$form, FormStateInterface $form_state) {
    $response = new AjaxResponse();

    // Get the filename from form state.
    $filename = $form_state->get('filter_filename');

    // Get the media ID from the original media item.
    // @todo needs to work pass mid for auto-selection.
    $original_mid = $form_state->get('filter_mid');

    // Set both cookies for filename and mid.
    $this->setFilenameFilterCookie($form_state, $filename, $original_mid);

    // Get current media library state.
    $state = $this->getMediaLibraryState($form_state);

    // Force a rebuild.
    $state->set('_media_library_form_rebuild', TRUE);

    // Build the updated library content.
    $library_content = $this->libraryUiBuilder->buildUi($state);

    // Replace the current form with the library content.
    $response->addCommand(
          new ReplaceCommand(
              '#media-library-add-form-wrapper',
              $library_content
          )
      );

    // Add command to trigger view refresh.
    $response->addCommand(
          new InvokeCommand(
              '.view-media-library',
              'trigger',
              ['RefreshView']
          )
      );

    return $response;
  }

  /**
   * Sets the cookies for media library filtering and selection.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param string|null $filename
   *   The filename of the media item.
   * @param string|null $media_id
   *   The unique media id of the media item.
   *
   * @return void
   *   Returns void
   */
  protected function setFilenameFilterCookie(FormStateInterface $form_state, ?string $filename = NULL, $media_id = NULL) {
    // Get the media library state which contains our field information.
    $state = $this->getMediaLibraryState($form_state);
    if (!$state) {
      return;
    }

    // Get the opener parameters.
    $opener_parameters = $state->getOpenerParameters();
    if (empty($opener_parameters['field_widget_id'])) {
      return;
    }

    $field_widget_id = $opener_parameters['field_widget_id'];

    // Set all our cookies using Pantheon's recommended STYXKEY prefix.
    if ($filename) {
      setcookie('STYXKEY_media_filename', $filename, time() + 300, '/');
    }

    if ($media_id) {
      setcookie('STYXKEY_media_select_id', $media_id, time() + 300, '/');
    }
    if ($field_widget_id) {
      setcookie('STYXKEY_media_field_id', $field_widget_id, time() + 300, '/');
    }

    $this->killSwitch->trigger();
  }

  /**
   * Check if duplicate detection is enabled for the current field widget.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return bool
   *   TRUE if duplicate detection is enabled, FALSE otherwise.
   */
  protected function isDuplicateDetectionEnabled(FormStateInterface $form_state) {
    // Get the media library state which contains our field information.
    $state = $this->getMediaLibraryState($form_state);
    if (!$state) {
      return FALSE;
    }

    // Get the opener parameters.
    $opener_parameters = $state->getOpenerParameters();
    if (empty($opener_parameters['field_widget_id'])) {
      return FALSE;
    }

    // Get the field name from the widget ID.
    $field_name = $opener_parameters['field_widget_id'];

    // Load the entity form display config.
    $entity_form_display = $this->entityTypeManager
      ->getStorage('entity_form_display')
      ->load($opener_parameters['entity_type_id'] . '.' . $opener_parameters['bundle'] . '.default');

    if (!$entity_form_display) {
      return FALSE;
    }

    // Get the widget component settings.
    $component = $entity_form_display->getComponent($field_name);

    // Check if duplicate detection is enabled in third party settings.
    return !empty($component['third_party_settings']['media_dupe_detect']['detect_duplicates']);
  }

  /**
   * {@inheritdoc}
   */
  protected function processInputValues(array $source_field_values, array $form, FormStateInterface $form_state) {
    // Clean up any existing duplicate detection cookies.
    setcookie('STYXKEY_media_select_id', '', time() - 3600, '/');
    setcookie('STYXKEY_media_field_id', '', time() - 3600, '/');
    $this->killSwitch->trigger();

    // Get uploaded files first.
    $files = parent::processInputValues($source_field_values, $form, $form_state);

    // Check if duplicate detection is enabled for this field.
    if (!$this->isDuplicateDetectionEnabled($form_state)) {
      return $files;
    }

    if (!empty($files)) {
      // Check first file for duplicates.
      $first_file = reset($files);
      $duplicates = $this->findDuplicatesByHash($first_file->getFileUri());

      if (!empty($duplicates)) {
        // Get the first duplicate.
        $first_duplicate = reset($duplicates);

        // Clean up the temporary file.
        $this->fileSystem->delete($first_file->getFileUri());

        // Store the duplicate ID in form state.
        $form_state->set('duplicate_media_id', $first_duplicate->mid);

        // Set a flag to redirect in the form submit.
        $form_state->set('redirect_to_duplicate', TRUE);

        // Return empty array to prevent further processing.
        return [];
      }
    }

    return $files;
  }

}
